open Ovm
open Utils
open Exceptions

let globaldescr = "OVM is a virtual machine for the Tree Language"
let heap_msg = Format.sprintf "Set the heap size (default : %d)" !heapsize
let stack_msg = Format.sprintf "Set the stack size (default : %d)" !stacksize

let options =
  let set_size (size : int ref) (x : int) : unit =
    size := Memory.sizeOfInt * div_ceiling x Memory.sizeOfInt
  in
  [
    ("-show-ast", Arg.Set show_ast, "Enables the printing of the AST");
    ("-heap-size", Arg.Int (set_size heapsize), heap_msg);
    ("-stack-size", Arg.Int (set_size stacksize), stack_msg);
    ("-trace", Arg.Set trace, "Print the trace execution");
    ("-lin", Arg.Set linearize, "Linearizes the program before evalution");
    ("-lir", Arg.Set lir_check, "Check if the program is low-level");
    ("-unsafe", Arg.Set unsafe, "Don't perform the static checks");
    ( "-32",
      Arg.Unit (fun () -> word_size := 32),
      "Evaluate the program using 32 bit semantics (default)" );
    ( "-64",
      Arg.Unit (fun () -> word_size := 64),
      "Evaluate the program using 64 bit semantics" );
  ]

let parse_args () =
  let set_prog s =
    if Sys.file_exists s then file := s
    else Terminal.error (Format.sprintf "%s : file not found" s)
  in
  Arg.parse_argv Sys.argv options set_prog globaldescr;
  if !file = "" then Terminal.error "no filename specified";
  !file

let () =
  try
    let file = parse_args () in
    if !trace then (
      Format.printf "@[Parsing ";
      Terminal.cyan_fprintf "%s@,@]" file);
    let prog = Fileparser.parse file in
    if not !unsafe then StaticChecks.go prog;
    if !lir_check then StaticChecks.linear_check_prog prog;
    let prog = if !linearize then Linearization.linearize prog else prog in
    if !show_ast then Format.printf "@,%a@,\n" Tree.print_prog prog;
    let out_buff = Buffer.create 10 in
    let out_channel = Format.formatter_of_buffer out_buff in
    let err_buff = Buffer.create 50 in
    let err_channel = Format.formatter_of_buffer err_buff in
    if !trace then
      Vmio.init ~out_channel ~err_channel ~trace:Format.std_formatter
        In_channel.stdin
    else Vmio.init ~out_channel ~err_channel In_channel.stdin;
    Option.iter (fun out -> Format.fprintf out "%!") !Vmio.trace;
    let end_state =
      let end_state, code =
        Eval.eval_prog ~heapsize:!heapsize ~stacksize:!stacksize prog
      in
      Option.iter (fun out -> Format.fprintf out "%!") !Vmio.trace;
      Vmio.trace_with_offset 0 (Format.sprintf "Exit with code : %d\n" code);
      end_state
    in
    (* Show temporaries *)
    Vmio.print_trace (Format.asprintf "\n%a\n%!" Memory.print_temp end_state);
    (* Show out and err *)
    Format.fprintf out_channel "%!";
    Format.fprintf err_channel "%!";
    Format.printf "%s%!" (Buffer.contents out_buff);
    Format.eprintf "%s%!" (Buffer.contents err_buff)
  with
  | Arg.Help msg | Arg.Bad msg -> Format.printf "%s" msg
  | LexingError msg | ParsingError msg | IllFormed msg | RuntimeError (msg, _)
    ->
      Terminal.error msg
