open Tree
open Memory
open Utils
open Exceptions

module Make (A : Arith) = struct
  (** [jump_to mem label] Jump to the label [label], using data in [mem]. *)
  let jump_to (mem : Memory.t) (from : location) (dest : label) =
    Vmio.trace_with_offset 0 (Format.sprintf "Jump to label %s\n" dest);
    raise (JumpException (dest, from, mem))

  (** [evalMove mem m e] Evaluate a move expression. Store the result of the
    evaluation of [e] in [m] :
    - If [m] is [TEMP t], store the result in the temporary [t].
    - If [m] is [MEM(x)], store the result in memory, at the address [x].
    - Otherwise, it raise a [VM_runtime_exception]. *)
  let rec evalMove (mem : Memory.t) (m : expr) (e : expr) : Memory.t =
    Vmio.trace_indent "eval move\n";
    let nmem, store, pp =
      match m.payload with
      | Temp t ->
          ( mem,
            (fun mem x -> Memory.rstore mem t x),
            Format.asprintf "%a" print_expr m )
      | Mem e1 -> (
          let nmem, r = eval_expr mem e1 in
          try
            let addr = Result.to_int r in
            ( nmem,
              (fun mem x -> Memory.mstore mem addr x),
              Format.sprintf "mem %d" addr )
          with Failure msg -> runtime_error msg m.loc nmem)
      | _ -> runtime_error "Left-value of move should be mem or temp" m.loc mem
    in
    let nnmem, x = eval_expr nmem e in
    Vmio.trace_dedent (Format.asprintf "move : %s <- %a\n" pp Result.print x);
    match x with
    | Int x -> store nnmem x
    | r ->
        runtime_error
          (Format.asprintf "Right-value of move : %a, not an integer"
             Result.print r)
          e.loc nnmem

  (** [evalCjump mem (op, e1, e2, t, f)] Evaluate the expressions [e1] and [e2]
    and compare their resluts, using [op] :
    - If the result is true, jump to the label [t].
    - If the result is false, jump to the label [f].
    - If one of those result is not an integer, raise a [RuntimeError]. *)
  and evalCjump (mem : Memory.t) (from : location)
      ((op, lexpr, rexpr, t, f) : relop * expr * expr * label * label) :
      Memory.t =
    Vmio.trace_indent "evalCjump\n";
    let nmem, l = eval_expr mem lexpr in
    let nnmem, r = eval_expr nmem rexpr in
    Vmio.trace_dedent
      (Format.asprintf "%s %a %a\n" (relop_to_string op) Result.print l
         Result.print r);
    match (l, r) with
    | Int left, Int right ->
        let left = A.of_int left in
        let right = A.of_int right in
        let res =
          match op with
          | Eq -> A.compare left right = 0
          | Neq -> A.compare left right <> 0
          | LT -> A.compare left right < 0
          | GT -> A.compare left right > 0
          | LE -> A.compare left right <= 0
          | GE -> A.compare left right >= 0
          | ULT -> A.unsigned_compare left right < 0
          | UGT -> A.unsigned_compare left right > 0
          | ULE -> A.unsigned_compare left right <= 0
          | UGE -> A.unsigned_compare left right >= 0
        in
        jump_to nnmem from (if res then t else f)
        (* mod 2^{32} *)
    | Int _, _ ->
        runtime_error
          (Format.asprintf "Right operand : %a, is not an integer" Result.print
             r)
          rexpr.loc nnmem
    | _, _ ->
        runtime_error
          (Format.asprintf "Left operand : %a, is not an integer" Result.print l)
          lexpr.loc nnmem

  (** [evalName loc mem name] Evaluate a label [name] and return its corresponding
    address using [mem]. *)
  and evalName ~loc (mem : Memory.t) (name : label) : Memory.t * Result.t =
    Vmio.trace_with_offset 0 (Format.sprintf "name %s" name);
    let r =
      try Memory.lfetch mem name
      with Not_found ->
        Vmio.print_trace "\n";
        runtime_error (Format.asprintf "label \"%s\" unknown" name) loc mem
    in
    Vmio.print_trace (Format.sprintf " -> %i\n" r);
    (mem, Int r)

  (** [evalBinop mem op e1 e2] Evaluate the expressions [e1] and [e2] and compute
    [op e1 e2]. *)
  and evalBinop (mem : Memory.t) (op : op) (lexpr : expr) (rexpr : expr) :
      Memory.t * Result.t =
    Vmio.trace_indent (Format.sprintf "eval op %s\n" (binop_to_string op));
    let nmem, left = eval_expr mem lexpr in
    let nnmem, right = eval_expr nmem rexpr in
    Vmio.trace_dedent
      (Format.asprintf "%s %a %a\n" (binop_to_string op) Result.print left
         Result.print right);
    match (left, right) with
    | Int left, Int right ->
        let left' = A.of_int left in
        let right' = A.of_int right in
        let res =
          match op with
          | Add -> A.add left' right'
          | Sub -> A.sub left' right'
          | Mul -> A.mul left' right'
          | Div -> A.div left' right'
          | Mod -> A.rem left' right'
          | And -> A.logand left' right'
          | Or -> A.logor left' right'
          | Xor -> A.logxor left' right'
          | LShift -> A.shift_left left' right
          | RShift -> A.shift_right_logical left' right
          | ARshift -> A.shift_right left' right
        in
        (nnmem, Int (A.to_int res))
        (* mod 2^{32} *)
    | Int _, _ ->
        runtime_error
          (Format.asprintf "Right operand : %a, is not an integer" Result.print
             right)
          rexpr.loc nnmem
    | _, _ ->
        runtime_error
          (Format.asprintf "Left operand : %a, is not an integer" Result.print
             left)
          lexpr.loc nnmem

  (** [evalMem mem e] Evaluate a [MEM] expression by evaluating [e] and returning
    the value at this address. *)
  and evalMem (mem : Memory.t) (e : expr) : Memory.t * Result.t =
    Vmio.trace_indent "eval mem\n";
    let nmem, m = eval_expr mem e in
    let r =
      try Memory.mfetch mem (Result.to_int m)
      with Failure msg -> raise (RuntimeError (msg, nmem))
    in
    Vmio.trace_dedent (Format.sprintf "mem %i = %i\n" (Result.to_int m) r);
    (nmem, Int r)

  (** [evalCall mem name args] Evaluate the call of a predefined function or user
    routine [name] with the memory state [mem] and the arguments [args]. *)
  and evalCall loc (mem : Memory.t) (f : label) (args : expr list) :
      Memory.t * Result.t =
    Vmio.trace_indent (Format.sprintf "evalCall %s\n" f);
    let mem, evalargs = List.fold_left_map eval_expr mem args in
    Vmio.change_offset (-2);
    Vmio.trace_indent
      (Format.asprintf "call %s %a\n" f
         (Format.pp_print_list ~pp_sep:Format.pp_print_space Result.print)
         evalargs);
    let nmem, res =
      try eval_routine f mem evalargs
      with Failure _ -> (
        try (Runtime.of_label f) loc mem (List.combine args evalargs)
        with Failure m -> raise (RuntimeError (m, mem)))
    in
    Vmio.trace_dedent
      (Format.asprintf "%s %a -> %a\n" f
         (Format.pp_print_list ~pp_sep:Format.pp_print_space Result.print)
         evalargs Result.print res);
    (nmem, res)

  (** [eval_routine name mem args] Evaluate the routine [name] with the memory
    state [mem] and the arguments [args].

    The call of [eval_routine name mem] raise a failure if [name] is not
    associated to a function in [mem], without the need of [args], allowing to
    do :
    {[
      let f =
        try eval_routine name mem
        with _ -> ...
      in f args
    ]} *)
  and eval_routine (name : label) (mem : Memory.t) :
      Result.t list -> Memory.t * Result.t =
    let body = Memory.cfetch mem name in
    fun (args : Result.t list) ->
      let args', _ =
        List.fold_left
          (fun (acc, i) v ->
            (SMap.add (Format.sprintf "i%d" i) (Result.to_int v) acc, i + 1))
          (SMap.singleton "rv" 0, 0)
          args
      in
      (* Create a buffer for routine's traces *)
      let trace = !Vmio.trace in
      let buff = Buffer.create 25 in
      let trace_buff = Format.formatter_of_buffer buff in
      if Option.is_some trace then Vmio.trace := Some trace_buff;
      let restore_trace_channel () =
        Format.fprintf trace_buff "%!";
        Option.iter
          (fun out ->
            if Vmio.remove_lazy_endline out then Format.fprintf out "@<0>\n";
            Format.fprintf out "%s" (String.of_seq (Buffer.to_seq buff));
            if List.mem trace_buff !Vmio.lazy_endline then
              Vmio.add_lazy_endline out)
          trace;
        Vmio.trace := trace
      in
      (* Save temporaries *)
      let temp = mem.temporaries.temp in
      let temporaries = { mem.temporaries with temp = args' } in
      (* Evaluate the routine *)
      let mem' =
        try eval_stmts_stack { mem with temporaries } None [ body ]
        with e ->
          restore_trace_channel ();
          raise e
      in
      let rv = Memory.rfetch mem' "rv" in
      (* Restore temporaries *)
      let temporaries = { mem'.temporaries with temp } in
      (* Print Routine traces *)
      restore_trace_channel ();
      ({ mem' with temporaries }, Int rv)

  (** [evalEseq mem s e] Evaluate the statement [s] and return the value of the
    evaluation of the expression [e]. *)
  and evalEseq (mem : Memory.t) (s : stmt) (e : expr) : Memory.t * Result.t =
    Vmio.trace_indent "evalEseq\n";
    (* TODO : use indices and array instead of assoc list *)
    let nmem = eval_stmts_stack mem (Some s) [ [ s ] ] in
    let nnmem, r = eval_expr nmem e in
    Vmio.trace_dedent (Format.asprintf "Eseq -> %a\n" Result.print r);
    (nnmem, r)

  (** [eval_expr mem e] Evaluate an expression [e] with the memory state [mem], by
    calling the function corresponding to the case, and return the result of the
    expression and the memory state after the evaluation.

    [eval_expr] is type [acc -> 'a -> acc * 'b] for [List.fold_left_map] *)
  and eval_expr (mem : Memory.t) (e : expr) : Memory.t * Result.t =
    match e.payload with
    | Cst x ->
        Vmio.print_trace (Format.sprintf "const %s\n" x);
        (mem, Int (A.of_string x |> A.to_int))
    | Name name -> evalName ~loc:e.loc mem name
    | Temp name -> (
        Vmio.print_trace (Format.sprintf "temp %s" name);
        try
          let v = rfetch mem name in
          Vmio.print_trace (Format.sprintf " = %i\n" v);
          (mem, Int v)
        with Not_found ->
          Vmio.print_trace "\n";
          runtime_error
            (Format.sprintf "No temporary named \"%s\"" name)
            e.loc mem)
    | Binop (op, e1, e2) -> evalBinop mem op e1 e2
    | Mem e -> evalMem mem e
    | Call ({ payload = Name f; _ }, args) -> evalCall e.loc mem f args
    | Call _ -> runtime_error "Call on non-name value" e.loc mem
    | Eseq (s, e) -> evalEseq mem s e

  (** [eval_stmt mem s] Evaluate a statement [s] with the memory state [mem], by
    calling the functio corresponding to the case, and return the memory state
    after the evaluation. *)
  and eval_stmt (mem : Memory.t) (s : stmt) : Memory.t =
    match s.payload with
    | Move (m, e) -> evalMove mem m e
    | Sxp e ->
        Vmio.trace_indent "eval sxp\n";
        let nmem, r = eval_expr mem e in
        Vmio.trace_dedent (Format.asprintf "sxp : %a\n" Result.print r);
        nmem
    | Jump ({ payload = Name label; _ }, _) -> jump_to mem s.loc label
    | Jump (e, _) -> runtime_error "Jump on non-label" e.loc mem
    | Cjump (op, left, right, t, f) ->
        evalCjump mem s.loc (op, left, right, t, f)
    | Seq ls -> evalSeq mem ls
    | Label l ->
        Vmio.trace_with_offset 0 (Format.sprintf "label %s\n" l);
        mem
    | Litteral (lab, str) ->
        Vmio.trace_with_offset 0 (Format.sprintf "label %s : \"%s\"\n" lab str);
        mem

  (** [evalSeq mem seq] Evaluate a sequence of statements [seq] and return the
    memory state after. *)
  and evalSeq (mem : Memory.t) (ls : stmt list) : Memory.t =
    Vmio.trace_indent "eval seq\n";
    let r = List.fold_left eval_stmt mem ls in
    Vmio.trace_dedent "end seq\n";
    r

  (** [eval_stmts_stack mem s stack] Evaluate a stack of sequences [stack], as
    defined in flow.ml. If a jump is catched, it will try to find its
    destination to evaluate it, using the memory state given :
    - If [s] is [None], it will search in the labels in top level
    - Otherwise, it will search in the labels of the corresponding [ESEQ]
    - If no destination is found, the jump is propagated *)
  and eval_stmts_stack (mem : Memory.t) (s : stmt option)
      (stack : stmt list list) : Memory.t =
    try
      match stack with
      | h :: t ->
          let nmem = List.fold_left eval_stmt mem h in
          List.fold_left
            (fun mem stmts ->
              Vmio.trace_dedent "end seq\n";
              List.fold_left eval_stmt mem stmts)
            nmem t
      | [] -> mem
    with JumpException (dest, loc, nmem) ->
      let stack, level =
        try Memory.jump_dest mem s dest
        with Not_found -> raise (JumpException (dest, loc, nmem))
      in
      if !jump_debug then
        Vmio.eprintf "Jump to@\n%a%!"
          (Format.pp_print_list
             ~pp_sep:(fun fmt () -> Format.fprintf fmt "@\nin@\n")
             (Format.pp_print_list
                ~pp_sep:(fun fmt () -> Format.fprintf fmt "@\n")
                print_stmt))
          stack;
      Vmio.set_offset ~number:level ~size:2;
      eval_stmts_stack nmem s stack

  let eval_prog ~(heapsize : int) ~(stacksize : int) (prog : stmt list) :
      Memory.t * int =
    try
      let mem = Memory.initialize ~heapsize ~stacksize prog in
      let main =
        try Memory.cfetch mem "main"
        with Failure _ -> raise (RuntimeError ("No routine main", mem))
      in
      try (eval_stmts_stack mem None [ main ], 0) with
      | ExitException (code, mem') -> (mem', code)
      | JumpException (label, loc, mem) ->
          runtime_error
            (Format.asprintf "Invalid jump destination : %s" label)
            loc mem
    with Preload.PreloadError (s, loc) -> ill_formed s loc
end

module Eval32 = Make (Int32)
module Eval64 = Make (Int64)

let eval_prog ~(heapsize : int) ~(stacksize : int) (prog : stmt list) :
    Memory.t * int =
  if !Utils.word_size = 32 then Eval32.eval_prog ~heapsize ~stacksize prog
  else Eval64.eval_prog ~heapsize ~stacksize prog
