open Tree
open Utils

(* two expressions commute if changing their evaluation order does not change
   the overall behaviour of the program : that is if we have memory equivalence
   and io equivalence.

   We have memory equivalence if no memory cell (either temporary or absolute
   address) is written in one of the two expression, and read or written in the
   other (i.e, read-read conflict do not matter)

   e.g:

   - 2, a // commute

   - a, (b+a) // commute

   - a, (b:=a) // commute

   - a, (a:=2; 42) // do not commute *)

(* fake location for ast parts that are inserted by linearization *)
let fake_loc = (Lexing.dummy_pos, Lexing.dummy_pos)

let loc a = {payload= a; loc= fake_loc}

(** builds the set of temporaries that appear in an expression *)
let rec temps_expr e =
  match e.payload with
  | Temp s -> SSet.singleton s
  | Binop (_, e1, e2) -> SSet.union (temps_expr e1) (temps_expr e2)
  | Eseq (stmt, e) -> SSet.union (temps_expr e) (temps_stmt stmt)
  | Cst _ | Name _ -> SSet.empty
  | Mem e -> temps_expr e
  | Call (f_expr, args) ->
      List.fold_left
        (fun acc e -> SSet.union (temps_expr e) acc)
        (temps_expr f_expr) args

(** builds the set of temporaries that appear in a statement *)
and temps_stmt s =
  match s.payload with
  | Move (e1, e2) -> SSet.union (temps_expr e1) (temps_expr e2)
  | Sxp e -> temps_expr e
  | Jump (e, _labs) -> temps_expr e
  | Cjump (_r, e1, e2, _l1, _l2) -> SSet.union (temps_expr e1) (temps_expr e2)
  | Seq s ->
      List.fold_left (fun acc x -> SSet.union acc (temps_stmt x)) SSet.empty s
  | Label _ -> SSet.empty
  | Litteral _ -> SSet.empty

(** builds the set of temporaries that appear in a program *)
let temps (prog : stmt list) : SSet.t =
  List.fold_left (fun acc x -> SSet.union acc (temps_stmt x)) SSet.empty prog

(** given a program p, returns a temporary generator that generates temporaries
    that are not in p *)
let fresh (prog : stmt list) : unit -> string =
  let temps = temps prog in
  let cpt = ref 0 in
  let rec loop () =
    let t = Format.sprintf "t%i" !cpt in
    incr cpt ;
    if SSet.mem t temps then loop () else t
  in
  loop

(** Store arguments in temporaries (if not already temporaries), and calls the
    function on those temporaries. Arguments are evaluated from left to right *)
let store_args temp_gen f_expr args =
  let prelude, args' =
    List.fold_left
      (fun (prelude, args) x ->
        match x.payload with
        | Temp _ -> (prelude, x :: args)
        | _ ->
            let t = loc (Temp (temp_gen ())) in
            (loc (Move (t, x)) :: prelude, t :: args) )
      ([], []) args
  in
  match (List.rev args', List.rev prelude) with
  | [], _ -> Call (f_expr, [])
  | [t], [] -> Call (f_expr, [t])
  | [t], [s] -> Eseq (s, loc (Call (f_expr, [t])))
  | args, prelude -> Eseq (loc (Seq prelude), loc (Call (f_expr, args)))

(** transformation of function calls into explicit call-by-value: f(e1,e2,...en)
    is rewritten into let t1 = e1 in let t2 = e2 in ... let tn = en in f
    (t1,t2,...,tn) *)
let rec cbv_expr temp_gen e =
  match e.payload with
  | Binop (op, e1, e2) ->
      let e1' = cbv_expr temp_gen e1 in
      let e2' = cbv_expr temp_gen e2 in
      {e with payload= Binop (op, e1', e2')}
  | Mem e -> {e with payload= Mem (cbv_expr temp_gen e)}
  | Call (f_expr, args) ->
      let f_expr' = cbv_expr temp_gen f_expr in
      let args' = List.map (cbv_expr temp_gen) args in
      {e with payload= store_args temp_gen f_expr' args'}
  | Eseq (stmt1, {payload= Eseq (stmt2, expr); _}) ->
      let expr' = cbv_expr temp_gen expr in
      let stmt1' = cbv_stmt temp_gen stmt1 in
      let stmt2' = cbv_stmt temp_gen stmt2 in
      loc (Eseq (loc (Seq [stmt1'; stmt2']), expr'))
  | Eseq (stmt, expr) ->
      let expr' = cbv_expr temp_gen expr in
      let stmt' = cbv_stmt temp_gen stmt in
      {e with payload= Eseq (stmt', expr')}
  (* Constants, litterals and temporaries stay the same *)
  | _ -> e

and cbv_stmt temp_gen s =
  match s.payload with
  | Move (e1, e2) ->
      let e1' = cbv_expr temp_gen e1 in
      let e2' = cbv_expr temp_gen e2 in
      loc (Move (e1', e2'))
  | Sxp e -> loc (Sxp (cbv_expr temp_gen e))
  | Jump (e, labs) -> loc (Jump (cbv_expr temp_gen e, labs))
  | Cjump (r, e1, e2, l1, l2) ->
      let e1' = cbv_expr temp_gen e1 in
      let e2' = cbv_expr temp_gen e2 in
      loc (Cjump (r, e1', e2', l1, l2))
  | Seq s -> loc (Seq (List.map (cbv_stmt temp_gen) s))
  (* labels and litterals stay the same *)
  | _ -> s

(** Sequence flattening *)
let rec flatten_seq l =
  List.fold_left
    (fun acc s ->
      match s.payload with Seq s -> acc @ flatten_seq s | _ -> acc @ [s] )
    [] l

(** Main linearization function : Eseq and Calls are moved toward the root of
    the AST, with careful attention to evaluation order. When the evaluation
    order between two sub-trees might impact the result, we must preserve the
    values of the sub-expressions. Calls shoud be already flattened *)
let rec linearize_expr (temp_gen : unit -> label) e =
  match e.payload with
  | Binop (op, {payload= Eseq (s, e1); _}, e2) ->
      let e1' = linearize_expr temp_gen e1 in
      let s' = linearize_stmt temp_gen s in
      let e2' = linearize_expr temp_gen e2 in
      loc (Eseq (s', loc (Binop (op, e1', e2'))))
  | Binop (op, ({payload= Call (_, _); _} as e1), e2) ->
      let t = loc (Temp (temp_gen ())) in
      let s = loc (Move (t, e1)) in
      let e2' = linearize_expr temp_gen e2 in
      loc (Eseq (s, loc (Binop (op, t, e2'))))
  | Binop (op, e1, {payload= Eseq (s, e2); _}) ->
      let e1' = linearize_expr temp_gen e1 in
      let s' = linearize_stmt temp_gen s in
      let e2' = linearize_expr temp_gen e2 in
      if Region.conflicts_s_e s' e1' then
        let t = loc (Temp (temp_gen ())) in
        let s' = loc (Seq [loc (Move (t, e1')); s]) in
        loc (Eseq (s', loc (Binop (op, t, e2'))))
      else loc (Eseq (s', loc (Binop (op, e1', e2'))))
  | Binop (op, e1, ({payload= Call (_, _); _} as e2)) ->
      let t = loc (Temp (temp_gen ())) in
      let s = loc (Move (t, e2)) in
      let e1' = linearize_expr temp_gen e1 in
      loc (Eseq (s, loc (Binop (op, e1', t))))
  | Binop (op, e1, e2) ->
      let e1' = linearize_expr temp_gen e1 in
      let e2' = linearize_expr temp_gen e2 in
      {e with payload= Binop (op, e1', e2')}
  | Mem {payload= Eseq (s, e); _} ->
      loc
        (Eseq (linearize_stmt temp_gen s, loc (Mem (linearize_expr temp_gen e))))
  | Mem e -> {e with payload= Mem (linearize_expr temp_gen e)}
  | Eseq (stmt1, {payload= Eseq (stmt2, expr); _}) ->
      let expr' = linearize_expr temp_gen expr in
      let stmt1' = linearize_stmt temp_gen stmt1 in
      let stmt2' = linearize_stmt temp_gen stmt2 in
      {e with payload= Eseq (loc (Seq [stmt1'; stmt2']), expr')}
  | Eseq (stmt, expr) ->
      let expr' = linearize_expr temp_gen expr in
      let stmt' = linearize_stmt temp_gen stmt in
      {e with payload= Eseq (stmt', expr')}
  (* Calls, Constants, litterals and temporaries stay the same *)
  | _ -> e

and linearize_stmt (temp_gen : unit -> label) s =
  match s.payload with
  | Move (e1, {payload= Eseq (s, e2); _}) ->
      let e1' = linearize_expr temp_gen e1 in
      let e2' = linearize_expr temp_gen e2 in
      loc (Seq [linearize_stmt temp_gen s; loc (Move (e1', e2'))])
  | Move (e1, e2) ->
      let e1' = linearize_expr temp_gen e1 in
      let e2' = linearize_expr temp_gen e2 in
      {s with payload= Move (e1', e2')}
  | Sxp {payload= Eseq (s, e); _} ->
      loc
        (Seq [linearize_stmt temp_gen s; loc (Sxp (linearize_expr temp_gen e))])
  | Sxp e -> loc (Sxp (linearize_expr temp_gen e))
  | Jump (e, labs) -> loc (Jump (linearize_expr temp_gen e, labs))
  | Cjump (r, e1, e2, l1, l2) ->
      let e1' = linearize_expr temp_gen e1 in
      let e2' = linearize_expr temp_gen e2 in
      loc (Cjump (r, e1', e2', l1, l2))
  | Seq s -> loc (Seq (flatten_seq (List.map (linearize_stmt temp_gen) s)))
  (* labels and litterals stay the same *)
  | _ -> s

(** iterates f, starting with x, until we reach a fixpoint *)
let rec fix f x =
  let fx = f x in
  if fx = x then x else fix f fx

(** flattening of the sequences and calling linearize until a fixpoint is
    reached *)
let linearize p =
  let temp_gen = fresh p in
  let go p =
    let flatten_calls = List.map (cbv_stmt temp_gen) p in
    flatten_seq (List.map (linearize_stmt temp_gen) flatten_calls)
  in
  fix go p
