open Utils
open Tree

(* a reachable memory region is either computable and is a finite set of
   temporaries, or it's the Top element ie it is uncomputable (or too expensive
   to calculate) and could be anything. This is an over-approximation *)

type t = Exact of SSet.t | Top

let is_empty = function Exact s -> SSet.is_empty s | Top -> false

let meet r1 r2 =
  match (r1, r2) with
  | Top, Top -> true
  | Top, Exact s | Exact s, Top -> not (SSet.is_empty s)
  | Exact s1, Exact s2 -> not (SSet.is_empty (SSet.inter s1 s2))

(** builds the region that is read in an expression *)
let rec reads_expr e =
  match e.payload with
  | Temp s -> SSet.singleton s
  | Binop (_, e1, e2) -> SSet.union (reads_expr e1) (reads_expr e2)
  | Eseq (stmt, e) -> SSet.union (reads_expr e) (reads_stmt stmt)
  | Cst _ | Name _ -> SSet.empty
  | Mem _ | Call _ -> raise Exit

(** builds the set of temporaries that are read in a statement *)
and reads_stmt s =
  match s.payload with
  | Move ({payload= Temp _; _}, e2) -> reads_expr e2
  | Move _ -> raise Exit
  | Sxp e -> reads_expr e
  | Jump (e, _labs) -> reads_expr e
  | Cjump (_r, e1, e2, _l1, _l2) -> SSet.union (reads_expr e1) (reads_expr e2)
  | Seq s ->
      List.fold_left (fun acc x -> SSet.union acc (reads_stmt x)) SSet.empty s
  | Label _ -> SSet.empty
  | Litteral _ -> SSet.empty

(** builds the region that is written in an expression *)
let rec writes_expr e =
  match e.payload with
  | Temp _ -> SSet.empty
  | Binop (_, e1, e2) -> SSet.union (reads_expr e1) (reads_expr e2)
  | Eseq (stmt, e) -> SSet.union (reads_expr e) (reads_stmt stmt)
  | Cst _ | Name _ -> SSet.empty
  | Mem _ | Call _ -> raise Exit

(** builds theregion that is written in a statement *)
and writes_stmt s =
  match s.payload with
  | Move ({payload= Temp t; _}, e2) ->
      SSet.union (SSet.singleton t) (writes_expr e2)
  | Move _ -> raise Exit
  | Sxp e -> reads_expr e
  | Jump (e, _labs) -> reads_expr e
  | Cjump (_r, e1, e2, _l1, _l2) -> SSet.union (reads_expr e1) (reads_expr e2)
  | Seq s ->
      List.fold_left (fun acc x -> SSet.union acc (reads_stmt x)) SSet.empty s
  | Label _ -> SSet.empty
  | Litteral _ -> SSet.empty

let reads_expr e = try Exact (reads_expr e) with Exit -> Top

let reads_stmt e = try Exact (reads_stmt e) with Exit -> Top

let writes_expr e = try Exact (writes_expr e) with Exit -> Top

let writes_stmt e = try Exact (writes_stmt e) with Exit -> Top

let conflicts_e_e e1 e2 =
  let w_e1 = writes_expr e1 in
  let w_e2 = writes_expr e2 in
  meet w_e1 w_e2 || meet w_e1 (reads_expr e2) || meet w_e2 (reads_expr e1)

let conflicts_s_e s1 e2 =
  let w_s1 = writes_stmt s1 in
  let w_e2 = writes_expr e2 in
  meet w_s1 w_e2 || meet w_s1 (reads_expr e2) || meet w_e2 (reads_stmt s1)
