open Utils
open Exceptions
open Tree

(* Jumps validity *)

(** [add_label_scope cur_scope seq] Add to [cur_scope] all the labels
    in the sequence [seq] and its contained sequences.  labels at
    toplevel, i.e routine's entry point, are not considered valid to
    jump to
*)
let rec add_label_scope toplevel (scope : SSet.t) (stmts : stmt list) : SSet.t =
  let aux toplevel (acc : SSet.t) (s : stmt) =
    match s.payload with
    | Label lab -> if toplevel then acc else SSet.add lab acc
    | Seq stmts -> add_label_scope false acc stmts
    | _ -> acc
  in
  List.fold_left (aux toplevel) scope stmts

(** [valid_jump_seq scope seq] Check if the jumps in a sequence's AST [seq] are
    valid. *)
let rec valid_jump_seq (scope : SSet.t) (stmts : stmt list) =
  List.iter (valid_jump_stmt scope) stmts

(** [valid_jump_stmt scope stmt] Check if the jumps in a statement's AST [stmt]
    are valid. *)
and valid_jump_stmt (scope : SSet.t) s =
  match s.payload with
  | Jump ({ payload = Name lab; _ }, _l) ->
      if not (SSet.mem lab scope) then ill_formed "illegal jump" s.loc
  | Jump (e, l) ->
      if not (List.for_all (Fun.flip SSet.mem scope) l) then
        ill_formed "illegal jump" s.loc
      else valid_jump_expr scope e
  | Cjump (_op, e1, e2, t, f) ->
      if not (SSet.mem t scope || SSet.mem f scope) then
        ill_formed "illegal jump" s.loc;
      valid_jump_expr scope e1;
      valid_jump_expr scope e2
  | Seq l -> valid_jump_seq scope l
  | Sxp e -> valid_jump_expr scope e
  | Move (e1, e2) ->
      valid_jump_expr scope e1;
      valid_jump_expr scope e2
  | Label _ | Litteral _ -> ()

(** [valid_jump_expr scope expe] Check if the jumps in an expression's AST
    [expr] are valid. *)
and valid_jump_expr (scope : SSet.t) e =
  match e.payload with
  | Cst _ | Name _ | Temp _ -> ()
  | Binop (_, l, r) ->
      valid_jump_expr scope l;
      valid_jump_expr scope r
  | Mem e -> valid_jump_expr scope e
  | Call (f, a) ->
      valid_jump_expr scope f;
      List.iter (valid_jump_expr scope) a
  | Eseq (s, e) ->
      valid_jump_stmt (add_label_scope false scope [ s ]) s;
      valid_jump_expr scope e

(** binds a label to it's definition point. If the label is already
    bound then we raise an error. If the label is "end" and we are not
    at toplevel, we raise an error *)
let add l stmt toplevel map =
  if l = "end" then
    if not toplevel then
      ill_formed "label \"end\" must only be defined at toplevel" stmt
    else map
  else
    SMap.update l
      (function
        | None -> Some stmt
        | Some s ->
            raise
              (IllFormed
                 (Format.asprintf
                    "Ill-formed program, label %s has multiple definitions :@\n\
                     %a\n\n\
                    \ and \n\
                     %a"
                    l Errors.from_loc s Errors.from_loc stmt)))
      map

(** [labels_check_stmt stmt] Return the map from labels name to their
    definitions in a statement. *)
let rec labels_check_stmt toplevel map (stmt : stmt) =
  match stmt.payload with
  | Label l | Litteral (l, _) -> add l stmt.loc toplevel map
  | Move (e1, e2) | Cjump (_, e1, e2, _, _) ->
      labels_check_expr (labels_check_expr map e1) e2
  | Sxp e | Jump (e, _) -> labels_check_expr map e
  | Seq l -> List.fold_left (labels_check_stmt false) map l

(** [labels_check_expr map expr]  Return the map from labels name to their
    definitions in an expression [expr]. *)
and labels_check_expr map (expr : expr) =
  match expr.payload with
  | Cst _ | Name _ | Temp _ -> map
  | Binop (_, e1, e2) -> labels_check_expr (labels_check_expr map e1) e2
  | Mem e -> labels_check_expr map e
  | Call (f, a) -> List.fold_left labels_check_expr (labels_check_expr map f) a
  | Eseq (s, e) -> labels_check_expr (labels_check_stmt false map s) e

(** [labels_check_seq prog] Return the map from labels name to their
    definitions in a program. *)
let labels_check_prog (prog : stmt list) =
  List.fold_left (labels_check_stmt true) SMap.empty prog

(* Linearity *)

(** [linear_check_expr call seq e] Check if the expression [expr] is linear.
    - [call] is [true] if [expr] is in a call.
    - [seq] is [true] if [expr] is in a sequence. *)
let rec linear_check_expr (call : bool) (seq : bool) e =
  match e.payload with
  | Cst _ | Name _ | Temp _ -> ()
  | Binop (_, l, r) ->
      linear_check_expr call seq l;
      linear_check_expr call seq r
  | Mem e -> linear_check_expr call seq e
  | Call (_f, _a) ->
      (* Call is not in a sxp or a move stmt if we get to there *)
      ill_formed "call illegal in LIR" e.loc
  | Eseq _ -> ill_formed "eseq not allowed in LIR" e.loc

(** [linear_check_seq call seq stmts] Check if the sequence [stmts] is linear.
    - [call] is [true] if [expr] is in a call.
    - [seq] is [true] if [expr] is in a sequence. *)
and linear_check_seq (call : bool) (seq : bool) (stmts : stmt list) =
  let rec aux = function
    | [] -> ()
    | s :: t ->
        linear_check_stmt call seq s t;
        aux t
  in
  aux stmts

(** [linear_check_stmt call seq s t] Check if the statement [s] is linear, with
    [t] the next statements.
    - [call] is [true] if [expr] is in a call.
    - [seq] is [true] if [expr] is in a sequence. *)
and linear_check_stmt (call : bool) (seq : bool) (s : stmt) (_st : stmt list) =
  match s.payload with
  | Sxp { payload = Call (f, a); _ } ->
      (* Call is non-linear if nested in a call *)
      linear_check_expr true seq f;
      List.iter (linear_check_expr true seq) a
  | Move (dest, { payload = Call (f, a); _ }) ->
      (* Call is non-linear if nested in a call *)
      linear_check_expr true seq f;
      linear_check_expr true seq dest;
      List.iter (linear_check_expr true seq) a
  | Move (e1, e2) ->
      linear_check_expr call seq e1;
      linear_check_expr call seq e2
  | Sxp e -> linear_check_expr call seq e
  | Jump (e, _l) -> linear_check_expr call seq e
  | Cjump (_op, e1, e2, _t, _f) ->
      (* TODO: cjump must be followed by the false label *)
      linear_check_expr call seq e1;
      linear_check_expr call seq e2
  | Seq l ->
      (* Seq is non-linear if nested in a seq *)
      if seq then ill_formed "nested seq not allowed in LIR" s.loc;
      linear_check_seq call true l
  | Label _ | Litteral _ -> ()

(** [linear_check_prog prog] Check if the program [prog] is linear. *)
let linear_check_prog (prog : stmt list) =
  let rec aux (seq : bool) = function
    | [] -> ()
    | { payload = Label "end"; _ } :: t -> aux false t
    | { payload = Seq l; _ } :: t ->
        linear_check_seq false true l;
        aux true t
    | s :: t ->
        linear_check_stmt false seq s t;
        aux seq t
  in
  aux false prog

let go prog =
  ignore (labels_check_prog prog);
  valid_jump_seq (add_label_scope true SSet.empty prog) prog
