open Exceptions
open Result
open Memory

(** size_aux * Give the size of a string stored in memory * WARNING : The acc
    must be inititialized at 0 *)
let rec size_aux (mem : Memory.t) (addr : int) (acc : int) : int =
  let x =
    try Memory.mfetch mem addr
    with Failure msg -> raise (RuntimeError (msg, mem))
  in
  if x = 0 then acc else size_aux mem (addr + sizeOfInt) (acc + 1)

(** size * Predefined function size * Give the size of a string stored in memory *)
let size =
  Function.unary (fun mem addr -> (mem, Int (size_aux mem addr 0))) "size"

(** strconcat * Predefined function concat * Create a string in memory which is
    the concatenation of two other *)
let strconcat =
  let memcpy mem dest src size =
    if size > 0 then Memory.mstore_seq mem dest (Memory.mfetch_seq mem src size)
    else if size = 0 then mem
    else failwith (Format.sprintf "memcpy : size %d < 0" size)
  in
  Function.binary
    (fun mem addr1 addr2 ->
      try
        let size1 = size_aux mem addr1 0 in
        let size2 = size_aux mem addr2 0 in
        let size = Memory.sizeOfInt * (size1 + size2 + 1) in
        let mem', addr_res = Memory.mreserve mem size in
        let mem' = memcpy mem' addr_res addr1 size1 in
        let mem' = memcpy mem' (addr_res + (4 * size1)) addr2 size2 in
        (Memory.mstore mem' (addr_res + size - 1) 0, Int addr_res)
      with Failure msg -> raise (RuntimeError (msg, mem)))
    "concat"

(** strcmp * Predefined function strcmp * Compare the strings a and b. Return -1
    if a < b, 0 if equal, and 1 otherwise. *)
let strcmp =
  let rec strcmp_aux (mem : Memory.t) (addr1 : int) (addr2 : int) : int =
    try
      let left = Memory.mfetch mem addr1 in
      let right = Memory.mfetch mem addr2 in
      if left < right then -1
      else if right < left then 1
      else if left = 0 then 0
      else strcmp_aux mem (addr1 + sizeOfInt) (addr2 + sizeOfInt)
    with Failure msg -> raise (RuntimeError (msg, mem))
  in
  Function.binary
    (fun mem addr1 addr2 -> (mem, Int (strcmp_aux mem addr1 addr2)))
    "strcmp"

(** equals * Predefined function streq * Return 1 if the strings a and b are
    equal, 0 otherwise. Often faster than strcmp to test string * equality *)
let equals =
  let rec equals_aux (mem : Memory.t) (addr1 : int) (addr2 : int) : bool =
    try
      let left = Memory.mfetch mem addr1 in
      let right = Memory.mfetch mem addr2 in
      left = right
      && (left = 0 || equals_aux mem (addr1 + sizeOfInt) (addr2 + sizeOfInt))
    with Failure msg -> raise (RuntimeError (msg, mem))
  in
  Function.binary
    (fun mem addr1 addr2 ->
      (mem, Int (if equals_aux mem addr1 addr2 then 1 else 0)))
    "equals"

(** substring * Predefined function substring * Return a string composed of the
    characters of string starting at the first character (0 being * the origin),
    and composed of length characters (i.e., up to and including the character
    first * + length - 1) * Let size be the size of the string, the following
    assertions must hold: * - 0 <= first * - 0 <= length * - first + length <=
    size *)
let substring =
  Function.ternary
    (fun mem addr first length ->
      if first < 0 then
        raise
          (RuntimeError
             ( Format.sprintf "substring : second argument : %i is not positive"
                 length,
               mem ))
      else if length < 0 then
        raise
          (RuntimeError
             ( Format.sprintf
                 "substring : third argument :  length %i is not positive"
                 length,
               mem ))
      else
        let size = size_aux mem addr 0 in
        if first + length > size then
          raise
            (RuntimeError
               ( Format.sprintf
                   "substring : first + length > size : %i + %i > %i" first
                   length size,
                 mem ))
        else
          let nmem, addr' = Memory.mreserve mem ((size + 1) * sizeOfInt) in
          let substr =
            Memory.mfetch_seq nmem (addr + (sizeOfInt * first)) length
          in
          let nnmem = Memory.mstore nmem (addr' + length) 0 in
          (Memory.mstore_seq nnmem addr' substr, Int addr'))
    "substring"
