(* This file describes the validity testing mecanism of HIR files.

   HIR files can end with special comments, followed with a description what is
   expected at execution.

   /* expected output */ /* expected memory*/

   The output description is precisely what should be obtained on stdout, where
   each line is surrounded with the comment delimeters '/*' and '*/'

   The memory description is a list of cells, one per line of the form: /*
   cell:value */ where cell is a register and value an integer.

   The memory state obtained at the end of the execution must contain all the
   cells defined in the description and is allowed to contain other cells. *)

open Ovm

type test = {
  test_input : string;
  expected_output : string;
  memory_cells : (string * int) list;
}

let parse_cell s =
  match String.split_on_char ':' s with
  | [ address; value ] -> (
      try
        let ivalue = int_of_string (String.trim value) in
        (String.trim address, ivalue)
      with _ ->
        failwith
          (Format.asprintf
             "was expecting a numerical value in cell %s but got %s" s value))
  | _ -> failwith (Format.asprintf "wrong format in test entry : %s" s)

(* Test building *)
let gather filename : test option =
  let present = ref false in
  let ic = open_in filename in
  let rec loop take ((input, output, memory) as acc) =
    match input_line ic with
    | "/* test input */" ->
        present := true;
        let take x (input, output, memory) = (x :: input, output, memory) in
        loop take acc
    | "/* expected output */" ->
        present := true;
        let take x (input, output, memory) = (input, x :: output, memory) in
        loop take acc
    | "/* expected memory */" ->
        present := true;
        let take x (input, output, memory) = (input, output, x :: memory) in
        loop take acc
    | s ->
        if String.starts_with ~prefix:"/*" s then
          loop take (take (String.sub s 2 (String.length s - 4)) acc)
        else loop take acc
    | exception End_of_file ->
        close_in ic;
        (List.rev input, List.rev output, List.rev memory)
  in
  let input, output, mem = loop (fun _ acc -> acc) ([], [], []) in
  if !present then
    let test_input = String.concat "\n" input in
    let expected_output = String.concat "\n" output in
    let memory_cells = List.map parse_cell mem in
    Some { test_input; expected_output; memory_cells }
  else None

let memory_check (temp : string) (value : int) (mem : Cpu.registry) : bool =
  try
    let actual = Cpu.rfetch mem temp in
    if value <> actual then (
      Format.printf
        "memory mismatch: got \"%s : %i\" but was expecting \"%s : %i\"%!" temp
        actual temp value;
      false)
    else true
  with _ ->
    Format.printf
      "memory mismatch: got \"%s : none\" but was expecting \"%s : %i\"%!" temp
      temp value;
    false

let output_check (expected : string) (actual : string) : bool =
  if expected <> actual then (
    Format.printf "output mismatch: got \"%s\" but was expecting \"%s\" %!"
      actual expected;
    false)
  else true

(* actual checking *)
let check name mem out =
  let test = gather name in
  let out = String.of_seq (Buffer.to_seq out) in
  match test with
  | None ->
      Terminal.yellow_fprintf "no test found\n";
      true
  | Some { test_input = _; expected_output; memory_cells } ->
      let check_output = output_check expected_output out in
      let check_memory =
        List.for_all (fun (s, i) -> memory_check s i mem) memory_cells
      in
      check_output && check_memory

(* iterates (not recursively) the function 'f' over every file of the directory
   'dir'. *)
let iter_directory f dir =
  if Sys.file_exists dir then
    if Sys.is_directory dir then (
      let dir = Sys.readdir dir |> Array.map (( ^ ) (dir ^ "/")) in
      Array.sort compare dir;
      Array.iter f dir)
    else invalid_arg (dir ^ " is not a directory")
  else invalid_arg (dir ^ " does not exist")

(* evaluates a file *)
let run file lin =
  let prog = Fileparser.parse file in
  let prog = if lin then Linearization.linearize prog else prog in
  let input : In_channel.t =
    match gather file with
    | Some v ->
        let name, oc = Filename.open_temp_file "input_" "" in
        Out_channel.output_string oc v.test_input;
        Out_channel.close oc;
        In_channel.open_text name
    | None -> In_channel.stdin
  in
  let out_buff = Buffer.create 10 in
  let output = Format.formatter_of_buffer out_buff in
  let err_buff = Buffer.create 25 in
  let errors = Format.formatter_of_buffer err_buff in

  Vmio.init ~out_channel:output ~err_channel:errors input;
  let memory, _ =
    Eval.eval_prog ~heapsize:!Utils.heapsize ~stacksize:!Utils.stacksize prog
  in
  check file memory.Memory.temporaries out_buff

let expect_success file lin =
  try
    let b = run file lin in
    if b then Terminal.ok () else Terminal.ko ();
    b
  with e ->
    let msg = Printexc.to_string e in
    Format.printf "%s" msg;
    Terminal.ko ();
    false

let expect_error file lin =
  try
    ignore (run file lin);
    Format.printf "file %s was expected to provoke an error" file;
    Terminal.ko ();
    false
  with
  | Exceptions.IllFormed _ | Exceptions.LexingError _
  | Exceptions.ParsingError _ | Exceptions.RuntimeError _ ->
      Terminal.ok ();
      true
  | e ->
      let msg = Printexc.to_string e in
      Format.printf "unexpected exception %s%!" msg;
      Terminal.ko ();
      false

let total = ref 0
let pass = ref 0

let launch lin =
  iter_directory
    (fun file ->
      Utils.file := file;
      if Filename.extension file = ".hir" || Filename.extension file = ".lir"
      then (
        Format.printf "testing %s %s: %!" (Filename.basename file)
          (if lin then "(linearized)" else "");
        incr total;
        if String.starts_with (Filename.basename file) ~prefix:"error" then (
          if expect_error file lin then incr pass)
        else if expect_success file lin then incr pass))
    (* tests are run from _/build/default/tests/*)
    "../../../treetests"

let _ =
  Printexc.record_backtrace true;
  Format.printf "\n--------------\n| Unit tests |\n--------------\n\n";
  Testmemory.test_memory ();

  Format.printf
    "\n-------------------\n| Black box tests |\n-------------------\n\n";
  launch false;
  launch true;
  Format.printf "%i/%i test passed\n" !pass !total
